determine_missing_edge_lists <- function(edge_list,total_node_count,method_conn='all',method_disconn='all',
                             threshold_values=NULL,cut_off_conn=NULL,cut_off_disconn=NULL,
                             edge_count_conn=NULL,edge_count_disconn=NULL){
  #create a list of edges (tuples: target node, source node) for which to compute the missing edge probabilities, node indices
  # will be defined by the order of occurrence of the nodes in the edge list (not by the indices given in the edge list)
  # returns two entities: (i) list of edges between connected nodes, i.e. nodes which occur in the edge list, and (ii)
  # a list of edges for which target or source node is a disconnected node
  # 
  # INPUT
  #---------- 
  #edge_list - list of edges of the network, as defined in R
  #total_node_count - integer, number of nodes in the network (including those not occurring in the edge list)
  #threshold_values - symmetric square matrix of numeric values, usually correlation or pvalue matrix, 
  #               which gives the values at which to determine a threshold, node-by-node matrix,
  #               the column and row is considered to coincide with the node names from the edge list;
  #               ATTENTION: Edges with largest values in this matrix are preferred. Therefore, use 1-pvalue to sort for most 
  #               significant missing edges!
  #method_conn - method to use for the missing edge list of connected nodes, one of 'all', 'threshold' or 'edge_count'
  #method_disconn - method to use for the missing edge list of disconnected nodes, one of 'all', 'threshold' or 'edge_count', will usually be 'all' because we
  # have only quite few entries
  #cut_off_conn - numeric value, required if method_conn=='threshold': which cut-off value to use for edges between connected node edges (threshold_values>cut_off_conn)
  #cut_off_disconn - numeric value, required if method_disconn=='threshold': which cut-off value to use for edges connecting at least 1 disconnected node (threshold_val>cut_off_disconn)
  #edge_count_conn - integer value, required if method_conn=='edge_count', number of edges between connected nodes to return, ranked by threshold_values
  #edge_count_disconn - integer value, required if method_disconn=='edge_count', number of edges between disconnected nodes to return, ranked by threshold_values
  #
  #Output
  #----------------------
  #miss_list_conn: column 1: source node, column 2: target node of the edges; both are connected nodes only
  #edge_list_disconn: column 1: source node (conn node!), column 2: target node of the edges
  #these lists are directly usable for an SBM in Python established using the edge_list as input (i.e. first 
  #mention of a node corresponds to its index)
  #
  ####examples
  #-----------------
  #el_metab <- read.table('example_files/180905_metabbud_erneg_spear_BH_005_el.csv',
  #             sep=',',header=TRUE)
  #total_node_count <- 162
  #missing_edge_lists <- determine_missing_edge_lists(edge_list=el_metab,total_node_count=total_node_count)
  #
  ##example with thresholding
  #el_metab <- read.table('example_files/180905_metabbud_erneg_spear_BH_005_el.csv',
  #             sep=',',header=TRUE)
  #total_node_count <- 162
  #corr_metab <- readRDS('example_files/180502_metabbud_erneg_spear_corr.rds')
  #missing_edge_lists_threshold <- determine_missing_edge_lists(edge_list=el_metab,
  #                                     total_node_count=total_node_count,
  #                                     threshold_values=abs(corr_metab),
  #                                     method_conn='threshold',
  #                                     method_disconn='threshold',
  #                                     cut_off_conn=0.3,
  #                                     cut_off_disconn=0.1)
  ##For usage with Python edge confidence score/edge probability detection, save these lists as tables.
  #write.table(missing_edge_lists_threshold[[1]],"metab_example_missing_edge_list_connected_nodes.txt",sep="\t",row.names=F,
  #col.names=F,quote=F)
  #write.table(missing_edge_lists_threshold[[2]],"metab_example_missing_edge_list_disconnected_nodes.txt",sep="\t",row.names=F,
  #col.names=F,quote=F)
  #
  ###attention: we need to give absolute values in order for thresholding to favor edges with strong interaction (neg & pos)!
  #
  ##example with exact number of edges
  ##in order to get the same as above, we should enter 66 for disconnected and 194 for connected nodes
  #missing_edge_lists_threshold_count  <- determine_missing_edge_lists(edge_list=el_metab,
  #                                     total_node_count=total_node_count,
  #                                     threshold_values=abs(corr_metab),
  #                                     method_conn='edge_count',
  #                                     method_disconn='edge_count',
  #                                     edge_count_conn=194,
  #                                     edge_count_disconn=66)
  
  #it seems we have altered the edge lists to also start with indices from 0
  #revert that first
  edge_list[,1:2] <- edge_list[,1:2]+1
  
  #determine indices of connected nodes
  edge_list_st <- c(t(edge_list[,1:2])) #make to one vector
  edge_list_st <- edge_list_st[-which(duplicated(edge_list_st))] #keep only first entry of each node name
  conn_node_count <- length(edge_list_st) #how many connected nodes are there
  #important note: edge_list_st are only the indices necessary to deal with e.g. correlation tabs from R. All indices
  #of connected nodes will be 0:(conn_node_count-1) in Python
  #edge_list_st[1] coincides with node 0 of the SBM,
  #edge_list_st[2] coincides with node 1 of the SBM
  
  
  #determine edge list for connected nodes
  if (method_conn=='all'){ #if all conn edge pairs shall be given
    #save the existing edges into a matrix of size total_node_count
    miss_edge_available <- matrix(0,ncol=total_node_count,nrow=total_node_count)#matrix saving which edges are available and which not
    miss_edge_available[as.matrix(edge_list[,1:2])] <- -1 #set -1 to not-used edges
    #set edges to any disconnect edge to -1
    miss_edge_available[(1:total_node_count)[-edge_list_st],] <- -1
    #set diagonal to -1
    diag(miss_edge_available) <- -1
    #add the transposed to the matrix to get it all to -1 and below which edges should not be considered
    miss_edge_available <- miss_edge_available+t(miss_edge_available)
    #set all entries below the diagonal to -1
    miss_edge_available[lower.tri(miss_edge_available)] <- -1
    edge_list_conn <- which(!(miss_edge_available<0),arr.ind=TRUE) #use indices which haven't gotten some -1 values
    #these are the indices in R order. We need to retransform it to Python indices
    indvec <- rep(0,total_node_count)
    indvec[edge_list_st] <- 1:conn_node_count
    edge_list_conn <- cbind(indvec[edge_list_conn[,1]],indvec[edge_list_conn[,2]])
  }
  
  if (method_conn=='threshold'){
    #save the existing edges into a matrix of size total_node_count
    miss_edge_available <- matrix(0,ncol=total_node_count,nrow=total_node_count)#matrix saving which edges are available and which not
    miss_edge_available[as.matrix(edge_list[,1:2])] <- -1 #set -1 to not-used edges
    #set edges to any disconnected node to -1
    miss_edge_available[(1:total_node_count)[-edge_list_st],] <- -1
    #set diagonal to -1
    diag(miss_edge_available) <- -1
    #add the transposed to the matrix to get it all to -1 and below which edges should not be considered
    miss_edge_available <- miss_edge_available+t(miss_edge_available)
    #set all entries below the diagonal to -1
    miss_edge_available[lower.tri(miss_edge_available)] <- -1
    
    #now check the threshold_values matrix for the threshold values
    edge_list_conn <- which((!(miss_edge_available<0)) & #use indices which haven't gotten some -1 values
                              threshold_values>cut_off_conn, #threshold values should be bigger than cut_off
                              arr.ind=TRUE) #get indices
    #these are the indices in R order. We need to retransform it to Python indices
    indvec <- rep(0,total_node_count)
    indvec[edge_list_st] <- 1:conn_node_count
    edge_list_conn <- cbind(indvec[edge_list_conn[,1]],indvec[edge_list_conn[,2]])
  }
  
  if (method_conn=='edge_count'){
    #save the existing edges into a matrix of size total_node_count
    miss_edge_available <- matrix(0,ncol=total_node_count,nrow=total_node_count)#matrix saving which edges are available and which not
    miss_edge_available[as.matrix(edge_list[,1:2])] <- -1 #set -1 to not-used edges
    #set edges to any disconnect edge to -1
    miss_edge_available[(1:total_node_count)[-edge_list_st],] <- -1
    #set diagonal to -1
    diag(miss_edge_available) <- -1
    #add the transposed to the matrix to get it all to -1 and below which edges should not be considered
    miss_edge_available <- miss_edge_available+t(miss_edge_available)
    #set all entries below the diagonal to -1
    miss_edge_available[lower.tri(miss_edge_available)] <- -1
    
    thres_vals <- threshold_values[which(!(miss_edge_available<0),arr.ind=TRUE)] #vector of threshold values
    #determine the cut_off which corresponds to the edge_count
    ord_thres_vals <- order(thres_vals,decreasing=TRUE)
    edge_list_conn <- which(!(miss_edge_available<0),arr.ind=TRUE)[ord_thres_vals[1:edge_count_conn],]
    #these are the indices in R order. We need to retransform it to Python indices
    indvec <- rep(0,total_node_count)
    indvec[edge_list_st] <- 1:conn_node_count
    edge_list_conn <- cbind(indvec[edge_list_conn[,1]],indvec[edge_list_conn[,2]])
  }
  
  if (method_disconn=='all') { #if all disconn edge pairs shall be given
    edge_list_disconn=NULL
    if (total_node_count>length(edge_list_disconn)) { #if we have disconnected nodes
      edge_list_disconn <- cbind(1:conn_node_count,rep(conn_node_count+1,times=conn_node_count)) #all pairs between a disconnected and a connected node
      if (total_node_count-conn_node_count>1){ # if we have more than 1 disconnected node
        edge_list_disconn <- rbind(edge_list_disconn,c(conn_node_count+1,conn_node_count+2))
      }
    } 
  }
  if (method_disconn=='threshold'){# if we only use those bigger than a threshold
    edge_list_disconn=NULL
    if (total_node_count>length(edge_list_disconn)) { #if we have disconnected nodes
      #list of indices of disconnected nodes
      disconn_nodes <- (1:total_node_count)[-edge_list_st] #remove the "connected nodes" from all nodes
      #only look at the columns of the threshold matrix for the connected entries
      #row-wise maximal value for the threshold matrix, restricted to the combinations of connected and disconnected nodes
      #determine the maximal value in the threshold values matrix for each connected node over all disconnected nodes
      thres_max_vals_conn <- do.call(pmax,as.data.frame(threshold_values[edge_list_st,disconn_nodes]))#max over a row, for each row
      #each entry gives which connected node shall I use
      edge_list_disconn <- cbind(which(thres_max_vals_conn>cut_off_disconn), #connected nodes (do not use indices from R but from Python)
                                 rep(conn_node_count+1,times=sum(thres_max_vals_conn>cut_off_disconn))) # the first disconn node
      if (total_node_count-conn_node_count>1){ # if we have more than 1 disconnected node: add the edge between two disconnected nodes
        edge_list_disconn <- rbind(edge_list_disconn,c(conn_node_count+1,conn_node_count+2))
      }
    }
  }
  if (method_disconn=='edge_count'){#if we use a certain number (count) of edges, edges ranked by threshold_values
    edge_list_disconn=NULL
    if (total_node_count>length(edge_list_disconn)) { #if we have disconnected nodes
      #list of indices of disconnected nodes
      disconn_nodes <- (1:total_node_count)[-edge_list_st] #remove the "connected nodes" from all nodes
      #only look at the columns of the threshold matrix for the connected entries
      #row-wise maximal value for the threshold matrix, restricted to the combinations of connected and disconnected nodes
      #determine the maximal value in the threshold values matrix for each connected node over all disconnected nodes
      thres_max_vals_conn <- do.call(pmax,as.data.frame(threshold_values[edge_list_st,disconn_nodes]))
      #sort the entries in descedning order, take the edge_count first
      order_thres_max_vals_conn <- order(thres_max_vals_conn,decreasing = TRUE)
      edge_list_disconn <- cbind(order_thres_max_vals_conn[1:edge_count_disconn],
                                 rep(conn_node_count+1,times=edge_count_disconn))
      if (total_node_count-conn_node_count>1){ # if we have more than 1 disconnected node
        edge_list_disconn <- rbind(edge_list_disconn,c(conn_node_count+1,conn_node_count+2))
      }
    
    }
  }
  return(list(edge_list_conn-1,edge_list_disconn-1))#make the indices Python-compatible (starting at 0...)
}