
#distance measures


comp_reacterm_dist <- function(term1, term2, hier_tab, max_hier_depth){
  #find distance between two terms in a parent-child hierarchy given in hier_tab
  #
  #Input
  #---------------------
  #term1: string of one term
  #term2: string of 2nd term to compare
  #hier_tab: hierarchy file, given by a two-column format, first column gives the parents, second column the children
  #max_hier_depth: maximal depth of hierarchy (necessary to preallocate the space required)
  #
  #Output
  #----------------------
  #distance (level count from term1 plus term2 to nearest common ancestor term, minus 2 (to account for distance=0 for identical terms))
  #hierarchy level of common ancestor - reported as distance from the highest hierarchy level observed
  #level count from term 1
  #level count from term 2
  #
  #Example
  ##define hierachy file from Reactome
  #reac_hier <- read.table("/Users/kabaum/Desktop/fits SBM/biological annotation/ReactomePathwaysRelation.txt", sep="\t", header=FALSE, stringsAsFactors = FALSE)
  ##define the human terms only:
  #reac_hier_hsa <- reac_hier[grep("HSA",reac_hier$V1),]
  ##determine the distance between two terms
  #dist1 <- comp_reacterm_dist('R-HSA-1462054','R-HSA-1461957',reac_hier_hsa, max_hier_depth = 15)
  #
  ##one could also add a toplevel node to the 26 disconnected topnodes of reactome
  #add nodes to connect all 26 topnodes
  #reac_hier_hsa_topnodes <- rbind(reac_hier_hsa, 
  #                                data.frame(V1='R-HSA-TOPNODE', 
  #                                           V2=unique(reac_hier_hsa$V1[which(!(reac_hier_hsa$V1 %in% reac_hier_hsa$V2))])))
  ##get the distance between two terms
  #comp_reacterm_dist('R-HSA-1462054','R-HSA-6803157',hier_tab=reac_hier_hsa, max_hier_depth=15)
  ##we would get Inf as distance if terms are not connected
  #comp_reacterm_dist('R-HSA-1462054','R-HSA-6803157',hier_tab=reac_hier_hsa_topnodes, max_hier_depth=15)
  ##no infinite values
  #
  #determine all hierarchies at once, then compare and derive distance
  par_t1 <- vector("list", max_hier_depth) #list of parental terms
  par_t1[[1]] <- term1
  par_t2 <- vector("list", max_hier_depth) #list of parental terms
  par_t2[[1]] <- term2
  
  hier_lev  <- 1
  while ((hier_lev<max_hier_depth) & length(par_t1[[hier_lev]])>0) {
    hier_lev <- hier_lev+1
    par_t1[[hier_lev]] <- hier_tab[hier_tab[,2] %in% par_t1[[hier_lev-1]],1]
  }
  hier_lev  <- 1
  while ((hier_lev<max_hier_depth) & length(par_t2[[hier_lev]])>0) {
    hier_lev <- hier_lev+1
    par_t2[[hier_lev]] <- hier_tab[hier_tab[,2] %in% par_t2[[hier_lev-1]],1]
  }
  
  #check for term2 in term1
  anc_deg1 <- Inf
  anc_deg2 <- Inf
  hier_lev_ov <- Inf
  #go first through all parents from term 1, then through all parents from term2
  for (i in 1:max_hier_depth){
    overlap_calc <- unlist(lapply(par_t1,function(x){sum(par_t2[[i]] %in% x)})) #account for the fact that we might have multiple parents
    if (sum(overlap_calc)>0) {#then we have found an overlap
      anc_deg2=i
      anc_deg1=min(which(overlap_calc>0))
      hier_lev_ov <- hier_lev-anc_deg2 #number of parental terms for par_t2 minus found overlap 
      #print(overlap_calc)
      break
    }
  }
  return(c(anc_deg1+anc_deg2-2,hier_lev_ov,anc_deg1,anc_deg2))
}


compute_distances_between_blocks <- function(reactome_list1, reactome_list2, hierarchy_tab,max_hier_level){
  #given two lists of reactome terms, compute all distances between those
  dist_vec <- c(sapply(reactome_list1, function(x){
    c(sapply(reactome_list2,function(y){
      comp_reacterm_dist(x,y,hierarchy_tab,max_hier_level)
    }))
  }))
  return(dist_vec)
}

compute_distances_within_blocks <- function(reactome_list, hierarchy_tab,max_hier_level){
  #given a list of reactome terms, compute all distances between non-identical terms
  #compute_distances_within_blocks(paste0("R-HSA-",protulr_erneg_sf_enrichRPA[[1]][[3]]$GO_IDs_cor[1][[1]]),hierarchy_tab = reac_hier_hsa_topnodes,max_hier_level = 15)
  ind_mat <- expand.grid(1:length(reactome_list),1:length(reactome_list))
  ind_mat <- ind_mat[ind_mat[,1]>ind_mat[,2],]
  dist_vec <- c(sapply(1:dim(ind_mat)[[1]],function(x){comp_reacterm_dist(reactome_list[ind_mat[x,1]],
                                                                          reactome_list[ind_mat[x,2]],
                                                                          hierarchy_tab,max_hier_level)}))
  return(dist_vec)
}



compute_dist_all_blocks <- function(enrich_tab,hier_tab=reac_hier_hsa_topnodes,max_hier_depth=15,infinite_replace=20,infinite_replace_hier_ca=11){
  #computes all distances between reactome terms in blocks and returns a data frame with the values
  #enrich_tab should be output from compute_enrichPathway_partition()
  #hier_tab: hierarchy table (from reactome, reactome distances are used)
  #example
  #dist_mat_protulrsf <- compute_dist_all_blocks(enrich_tab=protulr_erneg_sf_enrichRPA[[1]])
  block_comb <- expand.grid(1:length(enrich_tab[[3]]$GO_IDs),1:length(enrich_tab[[3]]$GO_IDs)) #depends on no. of blocks only
  #initialize the data frame
  dist_data_tab<- data.frame(block_comb=block_comb[block_comb$Var1>=block_comb$Var2,],dist_mean=NA,dist_median=NA, dist_max=NA,dist_min=NA, #distances
                             hier_lev_ca_mean=NA,hier_lev_ca_median=NA,hier_lev_ca_min=NA,hier_lev_ca_max=NA) #hierarchy levels of common ancestor
  #
  dist_data_tab$block_comb_name <- "between blocks"
  dist_data_tab$block_comb_name[dist_data_tab$block_comb.Var1==dist_data_tab$block_comb.Var2] <- "within blocks"
  for (i in 1:dim(dist_data_tab)[[1]]){
    #only if both blocks have terms enriched
    print(i)
    if ((enrich_tab[[3]]$GO_count_cor[dist_data_tab$block_comb.Var1[i]]>0) & 
        (enrich_tab[[3]]$GO_count_cor[dist_data_tab$block_comb.Var2[i]]>0)) { #only if both blocks have enriched terms
      if (dist_data_tab$block_comb_name[i] == "within blocks") {
        distvec <- compute_distances_within_blocks(
          paste0("R-HSA-",enrich_tab[[3]]$GO_IDs_cor[dist_data_tab$block_comb.Var1[i]][[1]]),
          hier_tab,max_hier_level=max_hier_depth)
      } else { # if not within but between blocks
        distvec <- compute_distances_between_blocks(
          paste0("R-HSA-",enrich_tab[[3]]$GO_IDs_cor[dist_data_tab$block_comb.Var1[i]][[1]]), 
          paste0("R-HSA-",enrich_tab[[3]]$GO_IDs_cor[dist_data_tab$block_comb.Var2[i]][[1]]),
          hier_tab,max_hier_level=max_hier_depth)
      }
      #distances between terms: the larger the more distant the terms are
      dists <- distvec[seq(1,length(distvec),by=4)]
      dists[is.infinite(dists)] <- infinite_replace  #infinite values are set to 20
      dist_data_tab$dist_mean[i] <- mean(dists)
      dist_data_tab$dist_median[i] <- median(dists)
      dist_data_tab$dist_min[i] <- min(dists)
      dist_data_tab$dist_max[i] <- max(dists)
      #hierarchy level of common ancestor - the larger the closer, the smaller the more distant (topnode) - originally, we have 11 levels at most (starting from 1 to 11)
      hier_lev_ca <- distvec[seq(2,length(distvec),by=4)]
      #we map that here to values between 0 and 10 (level 10: topnode (very distant common ancestor), level 0: in branch node common ancestor (=identical))
      hier_lev_ca[is.infinite(hier_lev_ca)] <- infinite_replace_hier_ca #infinite values set to 11
      dist_data_tab$hier_lev_ca_mean[i] <- 11-mean(hier_lev_ca)
      dist_data_tab$hier_lev_ca_median[i] <- 11-median(hier_lev_ca)
      dist_data_tab$hier_lev_ca_max[i] <- 11-min(hier_lev_ca) #we have to switch min and max this here since we multiply the result by -1
      dist_data_tab$hier_lev_ca_min[i] <- 11-max(hier_lev_ca)
    }
  }
  
  return(dist_data_tab)
}






comp_sbmblock_dist <- function(sbm_block1, sbm_block2,sbm_hierarchy){
  #computes the distance between two blocks of an SBM, similar as distance in parental hierarchy of Reactome
  #the maximal hierarchy level is inferred from the dimension (number of columns) of sbm_hierarchy
  #
  #INPUT
  #----------
  #sbm_block1: identity of block 1, list of 2 numbers [[1]] : hierarchy level a
  #             (i.e. column in which this block occurs in the sbm_hierarchy dataframe) & [[2]] identity (block number) b
  #sbm_block2: identity of block 2, format as in sbm_block1
  #sbm_hierarchy: dataframe, column i = hierarchy level i, (the higher the higher the column index), entries should denote the block number
  #               this is as the output from 180903_retrieve_bestfitting_blocks.R or the function 
  #               get_partition_biol_anno_allnodesfitted() from useful_functions.R
  #
  #OUTPUT
  #------------
  #(distance value, hierarchy level of nearest common ancestor block, distance of term1 to common ancestor, distance of term2 to common ancestor)
  #
  #determine maximal hierarchy level
  max_hier_depth <- dim(sbm_hierarchy)[[2]]
  #print(max_hier_depth)
  #determine all parent of term1 and term2
  #print(sbm_block1[[2]][1])
  #print(sbm_block1[[1]][1])
  block_anc1 <- sbm_hierarchy[which(sbm_hierarchy[,sbm_block1[[1]][1]]==sbm_block1[[2]][1])[1],sbm_block1[[1]][1]:max_hier_depth]
  #print(block_anc1)
  hier_lev1 <- sbm_block1[[1]][1]:max_hier_depth
  #print(hier_lev1)
  block_anc2 <- sbm_hierarchy[which(sbm_hierarchy[,sbm_block2[[1]][1]]==sbm_block2[[2]][1])[1],sbm_block2[[1]][1]:max_hier_depth]
  #print(block_anc2)
  hier_lev2 <- sbm_block2[[1]][1]:max_hier_depth
  #
  #compare the entries
  #check for term2 in term1
  anc_deg1 <- Inf
  anc_deg2 <- Inf
  #go first through all parents from term 1, then through all parents from term2
  for (i in 1:length(hier_lev2)){
    overlap_calc <- (block_anc1 == block_anc2[i]) & (hier_lev1 == hier_lev2[i])
    #print(overlap_calc)
    if (sum(overlap_calc)>0) {#then we have found an overlap
      anc_deg2=i
      anc_deg1=min(which(overlap_calc))
      break
    }
  }
  return(c(anc_deg1+anc_deg2-2,hier_lev2[anc_deg2],anc_deg1,anc_deg2))
}


compute_sbm_distances_between_blocks <- function(block_pair_indices,hier_level_blocks,enrich_tab,sbm_hierarchy){
  #compute distances between all given pairs of tabs
  #uses the function comp_sbmblock_dist() from 180912_reactome_hierarchy.R
  dist_blocks_sbm <- rep(NA,dim(block_pair_indices)[[1]]) #how many rows we have in there
  merge_hier_level <- rep(NA,dim(block_pair_indices)[[1]])
  for (i in 1:dim(block_pair_indices)[[1]]){
    dist_blocks <- comp_sbmblock_dist(list(hier_level_blocks,as.character(enrich_tab[[3]]$partition_no[block_pair_indices[i,1]])), #first block
                                      list(hier_level_blocks,as.character(enrich_tab[[3]]$partition_no[block_pair_indices[i,2]])), #second blocks
                                      sbm_hierarchy) #only use the distance
    print(dist_blocks)
    merge_hier_level[i] <- dist_blocks[2]
    dist_blocks_sbm[i] <- dist_blocks[1]
  }
  return(list(dist_blocks_sbm,merge_hier_level))
}