library(clusterProfiler)

compute_enrichOwnAnno_partition <- function(part_table, background_genes, 
                                            term_2_gene, 
                                            term_2_name=NULL, 
                                            min_GSS=2,
                                            max_GSS=500,
                                            pval_cutoff=0.05,
                                            qval_cutoff=0.2,
                                            padj_method="BH"
                                            
){
  #this function computes the enrichment (hypergeometric test) for the block annotations 
  #using a user-defined annotation, given in TERM2GENE 
  #
  #INPUT
  #-------------
  #part_table: 2-column object giving the entity names (first column) and their partition (2nd column); 
  #partition: numbers from 0 to n-1
  #background_genes: which genes are at all in the list
  #term_2_gene: data frame given the employed annotation: first column: term name, 2nd column: molecule ID
  #term_2_name=NULL: (optional) map enrichment terms to names
  #min_GSS: minimal size of gene set to employ, default: 2
  #max_GSS: maximal size of gene set to employ, default: 500
  #pval_cutoff: p-value cut-off (normally 0.01, 0.05)
  #qval_cutoff: corrected p-value cut-off
  #padj_method: method for multiple testing correction
  #
  #OUTPUT: list
  #----------------
  #output[[1]]: list of enrichment objects (enrich_list)
  #output[[2]]: data frame giving different statistics already summarized: (metadata_table)
  # gene_count: number of genes in group
  # used_gene_count: gene counts for each enrichment computation (how many are there annotated)
  # term_count: list of numbers of KEGG Pathways which are enriched
  # term_IDs: list of KEGG Pathway IDs enriched 
  #
  ##Example of usage
  ##-------------------
  # #import partition
  # metabbud_sf_bestpart <- readRDS("~/example_files/180912_metabbud_erneg_sf_bestpart.RDS")
  # #retrieve KEGG compound IDs
  # metab_map_2_KEGG <- read.table("~/enrichment_analysis/Budczies_metabolite_anno_KEGG_semi_manual.txt",
  # sep="\t",header=T,stringsAsFactors = F,quote="")
  # #map to KEGG compound IDs
  # metabbud_sf_bestpart_KEGGC <- character(162)
  # for (i in 1:162) {
  #  metabbud_sf_bestpart_KEGGC[i] <- metab_map_2_KEGG$Output.ID.x[metab_map_2_KEGG$metabolite==metabbud_sf_bestpart[i,1]]
  # }
  # #retrieve KEGG pathway ontology for the KEGG compound IDs from the metabolites in the file
  # KEGGPW_2_KEGGID <- read.table("~/enrichment_analysis/KEGGpathwayAnno2KEGGID_metaboliteBudczies2013.txt",
  # sep="\t")
  # #retrieve the names of the KEGG Pathway IDs
  # KEGGPW_2_NAME <- read.table("~/enrichment_analysis/KEGGpathwayAnno2NAME_metaboliteBudczies2013.txt",sep="\t")
  # #apply the function to the lowest level partition
  # metab_part_enrichment <- compute_enrichOwnAnno_partition(
  #                         part_table=cbind(metabbud_sf_bestpart_KEGGC,metabbud_sf_bestpart[,2]),
  #                         background_genes=unique(KEGGPW_2_KEGGID$KEGGID),
  #                         term_2_gene=KEGGPW_2_KEGGID,
  #                         term_2_name=KEGGPW_2_NAME)

  #check how big the output needs to be
  outp_vec <- unique(part_table[,2]) #take into account that not all partition numbers need to be assigned!
  outp_size <- length(outp_vec) #count how many different identifiers, i.e. groups we have; 
  enrich_list=vector("list", length=outp_size) #initialize object saving the enrichments
  metadata_table <- data.frame(partition_no=outp_vec,gene_count=numeric(outp_size),
                               used_gene_count=numeric(outp_size),
                               term_count=numeric(outp_size))
  metadata_table$GO_IDs <- vector("list",length=outp_size) #initialize separately
  print(outp_size)
  for (i in 1:outp_size){
    print(i)
    metadata_table$gene_count[i] <- sum(part_table[,2]==outp_vec[i])
    #print(metadata_table$gene_count[i])
    entity_list=as.character(part_table[part_table[,2]==outp_vec[i],1])
    enrich_l <- enricher(entity_list, 
                         pvalueCutoff = pval_cutoff,
                         pAdjustMethod = padj_method, 
                         qvalueCutoff = qval_cutoff, 
                         minGSSize = min_GSS,
                         maxGSSize = max_GSS,
                         universe=background_genes,
                         TERM2GENE=term_2_gene,
                         TERM2NAME=term_2_name
    )
    #we have to catch the case that no terms of size>min_GSS is matched to the gene IDs. That's the case sometimes.
    #it will eliminate that element from the list of enrich_list. Thus, we have to define enrich_l before...
    if (is.null(enrich_l)) {
      enrich_list[[i]] <- NA
      metadata_table$used_gene_count[i] <- NA
      metadata_table$term_count[i] <- 0
      metadata_table$term_IDs[i] <- 0
    } else {#if GO terms were properly tested and enrich_l is not NULL
      enrich_list[[i]] <- enrich_l
      metadata_table$used_gene_count[i] <- length(enrich_list[[i]]@gene)
      metadata_table$term_count[i] <- length(enrich_list[[i]]$ID)
      metadata_table$term_IDs[i] <- list(enrich_list[[i]]$ID)
    }
  }
  return(list(enrich_list,metadata_table))
}
