library(ReactomePA)
library(clusterProfiler)
library(org.Hs.eg.db)

compute_enrichPathway_partition <- function(part_table, background_genes, 
                                            org_species='human',
                                            min_GSS=10,
                                            max_GSS=500,
                                            pval_cutoff=0.05,
                                            qval_cutoff=0.2,
                                            padj_method="BH",
                                            keytype="ENTREZID"
){
  #this function computes the Reactome Pathway enrichment for the block annotations (despite that 
  #output reads as if using GO annotations!). Entity annotations will be converted to ENTREZIDs.
  #
  #INPUT
  #-------------
  #part_table: 2-column object giving the entity names (first column) and their partition (2nd column); partition: numbers from 0 to n-1
  #background_genes: which genes are at all in the list
  #org_species: Reactome of which species to use, default: "human"
  #min_GSS: minimal size of gene set to employ, default: 10
  #max_GSS: maximal size of gene set to employ, default: 500
  #pval_cutoff: p-value cut-off (normally 0.01, 0.05)
  #qval_cutoff: corrected p-value cut-off
  #padj_method: method for multiple testing correction
  #keytype: which keytypes are entered as entities. The employed function requires ENTREZIDs, 
  #if keytype is something else, it will be converted using clusterProfiler::bitr()
  #
  #OUTPUT: list
  #----------------
  #output[[1]]: list of enrichment objects (enrich_list)
  #output[[2]]: data frame giving different statistics already summarized: (metadata_table)
  # gene_count: number of genes in group
  # used_gene_count: gene counts for each enrichment computation (how many are there)
  # term_count: list of numbers of Reactome terms for enrichment
  # term_IDs: list of Reactome IDs enriched 
  #
  ##Example 
  ##-------------------
  ##import partition
  #
  ##apply the function
  # #import a partition file
  # protein_example_part <- readRDS("~/example_files/180912_protulr_erneg_sf_bestpart.RDS")
  # #find background genes
  # background_genes_ENTREZ <- bitr(protein_example_part[,1], fromType = "SYMBOL", toType = "ENTREZID",OrgDb="org.Hs.eg.db")$ENTREZID
  # #compute enrichment for all blocks
  # protein_enrichment <- compute_enrichPathway_partition(protein_example_part[,1:2],
  #   background_genes=background_genes_ENTREZ,
  #   keytype="SYMBOL")
  #
  #
  
  #check how big the output needs to be
  outp_vec <- unique(part_table[,2]) #take into account that not all partition numbers need to be assigned!
  outp_size <- length(outp_vec) #count how many different identifiers, i.e. groups we have; 
  enrich_list=vector("list", length=outp_size) #initialize object saving the enrichments
  metadata_table <- data.frame(partition_no=outp_vec,gene_count=numeric(outp_size),
                               used_gene_count=numeric(outp_size),
                               term_count=numeric(outp_size))
  metadata_table$term_IDs <- vector("list",length=outp_size) #initialize separately
  print(outp_size)
  for (i in 1:outp_size){
    print(i)
    metadata_table$gene_count[i] <- sum(part_table[,2]==outp_vec[i])
    #print(metadata_table$gene_count[i])
    entity_list=as.character(part_table[part_table[,2]==outp_vec[i],1])
    if (!(keytype=="ENTREZID")) {#if entities other than ENTREZID are given, we need to convert
      entity_list=bitr(entity_list, fromType = keytype, toType = "ENTREZID",OrgDb="org.Hs.eg.db")$ENTREZID
    }
    enrich_l <- enrichPathway(entity_list, #the gene list which to check
                              organism = org_species, #the annotation origin
                              pvalueCutoff = pval_cutoff,
                              pAdjustMethod = padj_method, 
                              qvalueCutoff = qval_cutoff, 
                              minGSSize = min_GSS,
                              maxGSSize = max_GSS,
                              universe=background_genes, #background genes
                              readable=F #convert entrez gene ids to gene names
    )
    #we have to catch the case that no Reactome terms of size>min_GSS is matched to the gene IDs. That's the case sometimes.
    #it will eliminate that element from the list of enrich_list. Thus, we have to define enrich_l before...
    if (is.null(enrich_l)) {
      enrich_list[[i]] <- NA
      metadata_table$used_gene_count[i] <- NA
      metadata_table$term_count[i] <- 0
      metadata_table$term_IDs[i] <- 0
    } else {#if Reactome terms were properly tested and enrich_l is not NULL
      enrich_list[[i]] <- enrich_l
      metadata_table$used_gene_count[i] <- length(enrich_list[[i]]@gene)
      metadata_table$term_count[i] <- length(enrich_list[[i]]$ID)
      metadata_table$term_IDs[i] <- list(enrich_list[[i]]$ID)
    }
  }
  return(list(enrich_list,metadata_table))
}