get_partition_biol_anno <- function(topmindl_list, 
                                                   block_list, 
                                                   biol_entities, 
                                                   edge_list,
                                                   is_hierarchical = FALSE,
                                                   max_hier_level = 'all'){ #by default we export all hierarchy levels
  #this function allows for obtaining the partitions and biological annotations associated with the best X fits 
  #(as given by block_list) from a certain datafile
  #it employs the order of the nodes as indicated by the edge list which was generated from network reduction
  #this function works for block_list from a planar (non-hierarchical) SBM, and for hiera
  #function requires data produced with read_mindl() or read_top_mindl() (equivalent) and read_block_levels()
  #THE FITS ARE FOR ALL NODES OF THE MODEL, NOT ONLY FOR THOSE FROM THE EDGE LIST
  #
  #INPUT
  #-----------
  #topmindl_list: vector of minimal description length for the fits, partitions will be sorted by that
  #block_list: blocks, sets of fits are in same size as topmindl_list, e.g. in the order as obtained from read_block_levels
  #biol_entities: vector of biological identifiers
  #edge_list: edge list, i.e. order of edges; this is important to identify which nodes have been assigned to which labels; should be the input
  #     to the python file for networks establishment to ensure identical ordering
  #is_hierarchical [optional]: whether we have a hierarchical SBM (default: FALSE)
  #max_hier_level [optional]: how many levels to export if we have a hierarchical model, currently only "all" is supported
  #
  #Output: 
  #----------
  #partition_list: data frame with the biological entity identifier as first column, and each partition of the fits (sorted by smallest mindl)
  #               in subsequent columns, hierarchies one next to the other
  #
  #
  #
  #Examples
  #---------------------
  #for block partitions as derived from read_block_levels()
  # all_part_biol_metab_example <- get_partition_biol_anno(topmindl_list=topmindl_metab_example,
  #   block_list=all_blocks_metab_example,
  #   biol_entities=metab_biol_entities,
  #   edge_list=read.table('180905_metabbud_erneg_spear_scalefree_el.csv',sep=",",header=T),
  #   is_hierarchical = TRUE,
  #   max_hier_level = 'all')
  #
  ##for a single block partition
  # singlepart_biol_metab <- get_partition_biol_anno(topmindl_list=list(data.frame(V1=1)), 
  #   block_list=all_blocks_metab_example[[1]], 
  #   biol_entities=metab_biol_entities, 
  #   edge_list=read.table('180905_metabbud_erneg_spear_scalefree_el.csv',sep=",",header=T),
  #   is_hierarchical = TRUE,
  #   max_hier_level = 'all')
  
  #make sure that node number in edge_list is node number in biol_entities (raw data)
  #the order of occurrence of the nodes in the edge list will be the order in the python network,
  #get order of nodes as occurring in python network from edge list (indices start at 0!!!)
  edge_list_st <- c(t(edge_list[,1:2])) #merge edge list entries to one column, rowwise
  edge_list_st <- edge_list_st[-which(duplicated(edge_list_st))] #remove all but first occurrence of a node number
  #add the node indices which are not in the edge lists
  edge_list_st <- c(edge_list_st,(0:(length(biol_entities)-1))[-(edge_list_st+1)])
  
  #bring the topmindl in same order as blocks
  sorted_topmindl_list <- matrix(unlist(lapply(topmindl_list,function(x) {sort(x$V1)})), #x$V1 is the vector of topmindls for a set of fits
                                 ncol=length(block_list), #number of sets of fits
                                 nrow=length(block_list[[1]]),#number of fits per set
                                 byrow=FALSE)
  
  #order of the fits by smallest topmindl (over all sets, within the sets blocks are already sorted)
  ordered_blocks <- arrayInd(order(sorted_topmindl_list),dim(sorted_topmindl_list))
  print(ordered_blocks)
  
  if (!is_hierarchical){ # if we have a planar model, we can do it very simply
    print("non-hierarchical setting")
    #initialize container
    partition_list=cbind(biol_entities,matrix(NA,ncol=length(unlist(topmindl_list)),nrow=length(biol_entities)))
    ##fill the container 
    for (i in 1:dim(ordered_blocks)[[1]]) {
      #establish the vector of the partition to add 
      partition_ass <- rep(NA, dim(partition_list)[[1]])
      #set all nodes to their according value (block assignment+1)
      partition_ass[edge_list_st+1] <- 
        unlist(block_list[[ordered_blocks[i,2]]][[ordered_blocks[i,1]]]+1)
      #attach to other blocks
      partition_list[,i+1] <- partition_ass
    }
    
  } else {#if we have a hierarchical model, we could/need to supply the partitions of the hierarchies and the names
    #the block structure is given by block_list[[i]][[j]][[k]] for set of fits i, fit j (jth best fit in fit set), hierarchy level k
    if (max_hier_level == 'all') {# if all solutions shall be exported
      #initialize list
      print("hierarchical setting")
      partition_list=cbind(biol_entities,matrix(NA,
                                                ncol=sum(sapply(1:dim(ordered_blocks)[[1]],function(x){
                                                  length(block_list[[ordered_blocks[x,2]]][[ordered_blocks[x,1]]])
                                                })),
                                                nrow=length(biol_entities)))
      #define names of the partition list
      colnames(partition_list) <- c("biol_entity",unlist(sapply(1:dim(ordered_blocks)[[1]],
                                                                function(x){paste0("fit",x,"_lev",
                                                                                   seq(1,length(block_list[[ordered_blocks[x,2]]][[ordered_blocks[x,1]]])))})))
      #fill the list
      runind <- 1 #biol entities go in column 1
      for (i in 1:dim(ordered_blocks)[[1]]) { #for each fit
        #establish the vector of the partition to add 
        print(i)
        partition_ass <- rep(NA, dim(partition_list)[[1]])
        #for the lowest level it's straightforward
        partition_ass[edge_list_st+1] <- 
          unlist(block_list[[ordered_blocks[i,2]]][[ordered_blocks[i,1]]][[1]]+1)
        
        #attach to other blocks
        runind <- runind+1
        partition_list[,runind] <- partition_ass
        
        #the following levels rely on the respective previous level
        #set all other nodes to their according value in each hierarchy level (block assignment+1)
        for (k in 2:length(block_list[[ordered_blocks[i,2]]][[ordered_blocks[i,1]]])) { #for each level above 1
          print(k)
          partition_ass[edge_list_st+1] <- 
            unlist(block_list[[ordered_blocks[i,2]]][[ordered_blocks[i,1]]][[k]]+1)[partition_ass[edge_list_st+1]] #employ the level from before
          
          #attach to other blocks
          runind <- runind+1
          partition_list[,runind] <- partition_ass
        }
      }
    }
  }
  
  return(partition_list)
}
